/*
 * File          : $Source: /home/erdmann/Demo/src/RCS/msg.c,v $
 * Updated       : $Author: erdmann $
 * Last Modified : $Date: 2000/11/25 19:25:53 $
 * State         : $State: Released $
 * Version       : $Revision: 1.3 $
 *
 * Copyright (C) 2000 Michael Erdmann
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program/code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, US   
 *
 * Introduction
 * ============
 *
 * This modules contains a small implementation for message 
 * passing between different processes via Unix message 
 * queues.
 *
 * 
 * History
 * =======
 * $Log: msg.c,v $
 * Revision 1.3  2000/11/25 19:25:53  erdmann
 * Release Version 1_1
 *
 * Revision 1.2  2000/11/25 19:23:46  erdmann
 * Release Version 1_0
 *
 * Revision 1.1  2000/11/25 14:12:42  erdmann
 * No comments
 *
 */
# include <stdio.h>
# include <malloc.h>
# include <sys/types.h>
# include <sys/ipc.h>
# include <sys/msg.h>
# include <sys/fcntl.h>    

#include  "msg.h"

#define MESSAGE_TYPE 1
// ========================================================================= //
//                        L O C A L    P R O C E D U R E S		     //
// ========================================================================= //

/*
 * check if the space is realy available in the 
 * buffer.
 */
PRIVATE Space_Free( Message msg, int length ) {
   return msg->bytes_stored + length < sizeof( msg->buffer );
}

PRIVATE Space_Available( Message msg, int length ) {
   return msg->next_read + length < sizeof( msg->buffer );
}


PRIVATE Fetch( Message msg, void *data, int length ) {
#  ifdef DEBUG
   printf("fetch, next_read=%d\n", msg->next_read );
#  endif

   memcpy( data, msg->buffer + msg->next_read, length);
   msg->next_read += length;
}

PRIVATE Insert( Message msg, void *data, int length ) {
#  ifdef DEBUG
   printf("insert, bytes_stored=%d\n", msg->bytes_stored );
#  endif
   memcpy( msg->buffer + msg->bytes_stored, data, length);
   msg->bytes_stored += length;
}



// ========================================================================= //
//                       P U B L I C   P R O C E D U R E S		     //
// ========================================================================= //

/* 
 * Create a new message  
 */
PUBLIC Message New_Message( int id ) {
   Message Result;
   
   if( (Result=malloc(sizeof(Message_Buffer)))!=NULL ) {
      Result->bytes_stored = 0;
      Result->next_read    = 0;
      Result->id           = id;
   }
   return Result;
}

/* 
 * Add an integer to the internal buffer
 */
PUBLIC Message Add_Integer( Message msg, int value ) {
   if( !Space_Free(msg, sizeof(int)) ) 
      return NULL;

   Insert( msg, &value, sizeof(int) );
   return msg;
}

/*
 * Add a string into the message buffer. The string end
 * '\0' is part of the send data as well. 
 */
PUBLIC Message Add_String( Message msg, char *string ) {
   int length;

   length = strlen(string)+1;
   if( !Space_Free(msg, sizeof(int)+length) )
      return NULL;

   Add_Integer(msg, length );
   Insert(msg, string, length );
}

/*
 * Retreive an integer from the message buffer
 */
PUBLIC int Get_Integer( Message msg ) {
   int  Result;

   if( !Space_Available( msg, sizeof(int)) )
      return -1;
   Fetch( msg, &Result, sizeof(int) );
   return Result;
} 

/*
 * Get a string from the message buffer 
 */
PUBLIC char *Get_String( Message msg ) {
   int  length;
   char *Result = NULL;
   
   length = Get_Integer(msg);

   if( length > 0 && Space_Available(msg, length) ) {
      if( (Result=malloc(length))!=NULL )
         Fetch(msg, Result, length );
   }
   return Result; 
}


/*
 * wait for a message to be recevied. If the message has been 
 * recevied, the message buffer is copied into the fixed sized
 * result buffer and returned to the caller.
 */
Message Wait_For_Message( int qid ) { 
   int            msgid;	  	// unix message qeue id
   int            RC;			// Return codes for os primitives
   Message        Result = NULL;	// result of this function
   unsigned       length;		// length of the message data
   struct msgbuf  *unix_message; 	// unix message data structure

   if( (msgid = msgget( qid, IPC_CREAT | O_EXCL | 0777 )) == -1 ) {
      return NULL;
   };
      
   length = sizeof(struct msgbuf )+sizeof(Message_Buffer);
   if( (unix_message=malloc( length ))==NULL )
      return NULL;

   RC = msgrcv(msgid, unix_message, length, MESSAGE_TYPE, 0 );
  
   if( RC != -1 ) {
      if( (Result=(Message)malloc(sizeof(Message_Buffer)))!=NULL) {
         memcpy( Result, unix_message->mtext, sizeof(Message_Buffer) );
      }
   }
   free( unix_message );

   return Result;                      
}

/*
 * Convert our message into a unix message data structure and
 * send id with the unix message id  MESSAGE_TYPE
 */
int Message_Send( int qid, Message msg ) {
   int           RC;	      		// return code for unix os primitives
   int           msgid;			// unix message queue identifier
   unsigned      length;		// lenght of the unix message 
   struct msgbuf *unix_message;		// unix message buffer
 
   if( (msgid = msgget( qid, IPC_CREAT | O_EXCL | 0777 )) == -1 ) {
      return -1;
   };

   length = sizeof( struct msgbuf ) + sizeof( Message_Buffer );

   if( (unix_message=malloc(length))!=NULL ) {
      unix_message->mtype = MESSAGE_TYPE;
      memcpy( unix_message->mtext, msg, sizeof(Message_Buffer) );   
      msgsnd( msgid, unix_message, length, 0);

      free( unix_message );
   }  
}
